@@ -1,5 +1,7 @@
 use crate::ffi::decimal as ffi;
 
+use serde::{Serialize, Deserialize};
+
 #[derive(Debug, Copy, Clone)]
 pub struct Decimal {
     pub(crate) inner: ffi::decNumber,
@@ -9,6 +11,9 @@ impl Decimal {
     /// Initialize a `Decimal` instance from a raw [`decNumber`] struct
     ///
     /// [`decNumber`]: crate::ffi::decimal::decNumber
+    ///
+    /// # Safety
+    /// `inner` must a be valid instance of `decNumber` struct
     #[inline(always)]
     pub unsafe fn from_raw(inner: ffi::decNumber) -> Self {
         Self { inner }
@@ -299,7 +304,8 @@ impl std::fmt::Display for Decimal {
     #[inline(always)]
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         unsafe {
-            let c_ptr = ffi::decimal_to_string(&self.inner);
+            let mut buf = Vec::with_capacity((self.inner.digits + 14) as _);
+            let c_ptr = ffi::decNumberToString(&self.inner, buf.as_mut_ptr());
             let c_str = std::ffi::CStr::from_ptr(c_ptr);
             let r_str = std::str::from_utf8_unchecked(c_str.to_bytes());
             f.write_str(r_str)
@@ -546,13 +552,13 @@ macro_rules! impl_try_into_int {
             impl std::convert::TryFrom<Decimal> for $t {
                 type Error = DecimalToIntError;
 
-                fn try_from(mut dec: Decimal) -> Result<Self, Self::Error> {
+                fn try_from(dec: Decimal) -> Result<Self, Self::Error> {
                     if !dec.is_int() {
                         return Err(DecimalToIntError::NonInteger)
                     }
                     unsafe {
                         let mut num = std::mem::MaybeUninit::uninit();
-                        let res = $f(&mut dec.inner, num.as_mut_ptr());
+                        let res = $f(&dec.inner, num.as_mut_ptr());
                         if res.is_null() {
                             Err(DecimalToIntError::OutOfRange)
                         } else {
