@@ -7,16 +7,20 @@
 //! See also:
 //! - [Indexes](https://www.tarantool.io/en/doc/latest/book/box/data_model/#indexes)
 //! - [Lua reference: Submodule box.index](https://www.tarantool.io/en/doc/latest/reference/reference_lua/box_index/)
+use std::io::Write;
 use std::os::raw::c_char;
 use std::ptr::null_mut;
 use std::mem::MaybeUninit;
 
+use num_derive::ToPrimitive;
 use num_traits::ToPrimitive;
+use serde::Serialize;
 
 use crate::error::{Error, TarantoolError};
 use crate::ffi::tarantool as ffi;
 use crate::tuple::{AsTuple, Tuple, TupleBuffer};
 use crate::tuple_from_box_api;
+use crate::util::NumOrStr;
 
 /// An index is a group of key values and pointers.
 pub struct Index {
@@ -81,16 +85,85 @@ pub enum IteratorType {
     Neighbor = 11,
 }
 
+////////////////////////////////////////////////////////////////////////////////
+// Builder
+////////////////////////////////////////////////////////////////////////////////
+
+pub struct Builder<'a> {
+    space_id: u32,
+    name: &'a str,
+    opts: IndexOptions,
+}
+
+macro_rules! define_setters {
+    ($( $setter:ident ( $field:ident : $ty:ty ) )+) => {
+        $(
+            #[inline(always)]
+            pub fn $setter(mut self, $field: $ty) -> Self {
+                self.opts.$field = Some($field.into());
+                self
+            }
+        )+
+    }
+}
+
+impl<'a> Builder<'a> {
+    /// Creates a new index builder with default options.
+    #[inline(always)]
+    pub fn new(space_id: u32, name: &'a str) -> Self {
+        Self {
+            space_id,
+            name,
+            opts: IndexOptions::default()
+        }
+    }
+
+    define_setters!{
+        index_type(r#type: IndexType)
+        id(id: u32)
+        unique(unique: bool)
+        if_not_exists(if_not_exists: bool)
+        parts(parts: Vec<Part>)
+        dimension(dimension: u32)
+        distance(distance: RtreeIndexDistanceType)
+        bloom_fpr(bloom_fpr: f32)
+        page_size(page_size: u32)
+        range_size(range_size: u32)
+        run_count_per_level(run_count_per_level: u32)
+        run_size_ratio(run_size_ratio: f32)
+        sequence(sequence: impl Into<SequenceOpt>)
+        func(func: String)
+    }
+
+    #[inline(always)]
+    pub fn part(mut self, part: impl Into<Part>) -> Self {
+        self.opts.parts.get_or_insert_with(|| Vec::with_capacity(8))
+            .push(part.into());
+        self
+    }
+
+    /// Create a new index using the current options.
+    #[cfg(feature = "schema")]
+    #[inline(always)]
+    pub fn create(self) -> crate::Result<Index> {
+        crate::schema::index::create_index(self.space_id, self.name, &self.opts)
+    }
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// IndexOptions
+////////////////////////////////////////////////////////////////////////////////
+
 /// List of options for new or updated index.
 ///
 /// For details see [space_object:create_index - options](https://www.tarantool.io/en/doc/latest/reference/reference_lua/box_space/create_index/).
-#[derive(Serialize)]
+#[derive(Default, Serialize, tlua::Push)]
 pub struct IndexOptions {
-    pub index_type: Option<IndexType>,
+    pub r#type: Option<IndexType>,
     pub id: Option<u32>,
     pub unique: Option<bool>,
     pub if_not_exists: Option<bool>,
-    pub parts: Option<Vec<IndexPart>>,
+    pub parts: Option<Vec<Part>>,
     pub dimension: Option<u32>,
     pub distance: Option<RtreeIndexDistanceType>,
     pub bloom_fpr: Option<f32>,
@@ -98,60 +171,112 @@ pub struct IndexOptions {
     pub range_size: Option<u32>,
     pub run_count_per_level: Option<u32>,
     pub run_size_ratio: Option<f32>,
-    pub sequence: Option<IndexSequenceOption>,
+    pub sequence: Option<SequenceOpt>,
     pub func: Option<String>,
     // Only for Tarantool >= 2.6
     // pub hint: Option<bool>,
 }
 
-impl Default for IndexOptions {
-    fn default() -> Self {
-        IndexOptions {
-            index_type: Some(IndexType::Tree),
-            id: None,
-            unique: Some(true),
-            if_not_exists: Some(false),
-            parts: Some(vec![IndexPart {
-                field_index: 1,
-                field_type: IndexFieldType::Unsigned,
-                collation: None,
-                is_nullable: None,
-                path: None,
-            }]),
-            dimension: Some(2),
-            distance: Some(RtreeIndexDistanceType::Euclid),
-            bloom_fpr: Some(0.05),
-            page_size: Some(8 * 1024),
-            range_size: None,
-            run_count_per_level: Some(2),
-            run_size_ratio: Some(3.5),
-            sequence: None,
-            func: None,
-            // Only for Tarantool >= 2.6
-            // hint: Some(true),
-        }
-    }
-}
+////////////////////////////////////////////////////////////////////////////////
+// SequenceOpt
+////////////////////////////////////////////////////////////////////////////////
+
+#[deprecated = "Use `index::SequenceOpt` instead"]
+/// Use [`SequenceOpt`] instead
+pub type IndexSequenceOption = SequenceOpt;
 
 /// Sequence option for new or updated index.
 ///
 /// For details see [specifying a sequence in create_index](https://www.tarantool.io/en/doc/latest/reference/reference_lua/box_schema_sequence/create_index/#box-schema-sequence-create-index).
-#[derive(Serialize)]
-pub enum IndexSequenceOption {
-    SeqId {
-        seq_id: u32,
-        field_index: Option<u32>,
-    },
-    SeqName {
-        seq_name: String,
-        field_index: Option<u32>,
-    },
-    True,
-    Empty,
+#[derive(Serialize, tlua::Push)]
+pub enum SequenceOpt {
+    IdAndField(SeqSpec),
+    AutoGenerated(bool),
 }
 
+impl SequenceOpt {
+    #[inline(always)]
+    pub fn auto() -> Self {
+        Self::AutoGenerated(true)
+    }
+
+    #[inline(always)]
+    pub fn none() -> Self {
+        Self::AutoGenerated(false)
+    }
+
+    #[inline(always)]
+    pub fn field(field: impl Into<NumOrStr>) -> Self {
+        Self::IdAndField(SeqSpec::field(field))
+    }
+
+    #[inline(always)]
+    pub fn id(id: impl Into<NumOrStr>) -> Self {
+        Self::IdAndField(SeqSpec::id(id))
+    }
+
+    #[inline(always)]
+    pub fn spec(s: SeqSpec) -> Self {
+        Self::IdAndField(s)
+    }
+}
+
+impl From<SeqSpec> for SequenceOpt {
+    #[inline(always)]
+    fn from(s: SeqSpec) -> Self {
+        Self::spec(s)
+    }
+}
+
+impl From<bool> for SequenceOpt {
+    #[inline(always)]
+    fn from(b: bool) -> Self {
+        Self::AutoGenerated(b)
+    }
+}
+
+#[derive(Serialize, tlua::Push)]
+pub struct SeqSpec {
+    id: Option<NumOrStr>,
+    field: Option<NumOrStr>,
+}
+
+impl SeqSpec {
+    #[inline(always)]
+    pub fn field(field: impl Into<NumOrStr>) -> Self {
+        Self {
+            id: None,
+            field: Some(field.into()),
+        }
+    }
+
+    #[inline(always)]
+    pub fn id(id: impl Into<NumOrStr>) -> Self {
+        Self {
+            id: Some(id.into()),
+            field: None,
+        }
+    }
+
+    #[inline(always)]
+    pub fn and_field(mut self, field: impl Into<NumOrStr>) -> Self {
+        self.field = Some(field.into());
+        self
+    }
+
+    #[inline(always)]
+    pub fn and_id(mut self, id: impl Into<NumOrStr>) -> Self {
+        self.id = Some(id.into());
+        self
+    }
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// IndexType
+////////////////////////////////////////////////////////////////////////////////
+
 /// Type of index.
-#[derive(Copy, Clone, Debug, Serialize)]
+#[derive(Copy, Clone, Debug, Serialize, tlua::Push)]
 pub enum IndexType {
     Hash,
     Tree,
@@ -160,7 +285,7 @@ pub enum IndexType {
 }
 
 /// Type of index part.
-#[derive(Copy, Clone, Debug, Serialize)]
+#[derive(Copy, Clone, Debug, Serialize, tlua::Push)]
 pub enum IndexFieldType {
     Unsigned,
     String,
@@ -175,30 +300,106 @@ pub enum IndexFieldType {
     Scalar,
 }
 
+////////////////////////////////////////////////////////////////////////////////
+// IndexPart
+////////////////////////////////////////////////////////////////////////////////
+
+#[deprecated = "Use `index::Part` instead"]
+pub type IndexPart = Part;
+
 /// Index part.
-#[derive(Serialize)]
-pub struct IndexPart {
-    pub field_index: u32,
-    pub field_type: IndexFieldType,
+#[derive(Serialize, tlua::Push)]
+pub struct Part {
+    pub field: NumOrStr,
+    pub r#type: Option<IndexFieldType>,
     pub collation: Option<String>,
     pub is_nullable: Option<bool>,
     pub path: Option<String>,
 }
 
-impl IndexPart {
-    pub fn new(fi: u32, ft: IndexFieldType) -> Self {
-        return IndexPart {
-            field_index: fi,
-            field_type: ft,
-            collation: None,
-            is_nullable: None,
-            path: None,
-        };
+macro_rules! define_setters {
+    ($( $setter:ident ( $field:ident : $ty:ty ) )+) => {
+        $(
+            #[inline(always)]
+            pub fn $setter(mut self, $field: $ty) -> Self {
+                self.$field = Some($field.into());
+                self
+            }
+        )+
     }
 }
 
+impl Part {
+    pub fn field(field: impl Into<NumOrStr>) -> Self {
+        Self {
+            field: field.into(),
+            r#type: None,
+            collation: None,
+            is_nullable: None,
+            path: None,
+        }
+    }
+
+    define_setters!{
+        field_type(r#type: IndexFieldType)
+        collation(collation: String)
+        is_nullable(is_nullable: bool)
+        path(path: String)
+    }
+
+    pub fn new(fi: impl Into<NumOrStr>, ft: IndexFieldType) -> Self {
+        Self::field(fi).field_type(ft)
+    }
+}
+
+impl From<&str> for Part {
+    #[inline(always)]
+    fn from(f: &str) -> Self {
+        Self::field(f.to_string())
+    }
+}
+
+impl From<String> for Part {
+    #[inline(always)]
+    fn from(f: String) -> Self {
+        Self::field(f)
+    }
+}
+
+impl From<u32> for Part {
+    #[inline(always)]
+    fn from(f: u32) -> Self {
+        Self::field(f)
+    }
+}
+
+impl From<(u32, IndexFieldType)> for Part {
+    #[inline(always)]
+    fn from((f, t): (u32, IndexFieldType)) -> Self {
+        Self::field(f).field_type(t)
+    }
+}
+
+impl From<(String, IndexFieldType)> for Part {
+    #[inline(always)]
+    fn from((f, t): (String, IndexFieldType)) -> Self {
+        Self::field(f).field_type(t)
+    }
+}
+
+impl From<(&str, IndexFieldType)> for Part {
+    #[inline(always)]
+    fn from((f, t): (&str, IndexFieldType)) -> Self {
+        Self::field(f.to_string()).field_type(t)
+    }
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// ...
+////////////////////////////////////////////////////////////////////////////////
+
 /// Type of distance for retree index.
-#[derive(Copy, Clone, Debug, Serialize)]
+#[derive(Copy, Clone, Debug, Serialize, tlua::Push)]
 pub enum RtreeIndexDistanceType {
     Euclid,
     Manhattan,
@@ -233,7 +434,7 @@ impl Index {
                 self.space_id,
                 self.index_id,
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
                 @out
             ]
         )
@@ -259,7 +460,7 @@ impl Index {
                 self.index_id,
                 iterator_type.to_i32().unwrap(),
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
             )
         };
 
@@ -292,7 +493,7 @@ impl Index {
                 self.space_id,
                 self.index_id,
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
                 @out
             ]
         )
@@ -309,23 +510,41 @@ impl Index {
     /// Returns a new tuple.
     ///
     /// See also: [index.upsert()](#method.upsert)
-    pub fn update<K, Op>(&mut self, key: &K, ops: &Vec<Op>) -> Result<Option<Tuple>, Error>
+    pub fn update<K, Op>(&mut self, key: &K, ops: &[Op]) -> Result<Option<Tuple>, Error>
     where
         K: AsTuple,
         Op: AsTuple,
+    {
+        let mp_encoded_ops = ops.iter().try_fold(
+            Vec::with_capacity(ops.len()),
+            |mut v, op| -> crate::Result<Vec<Vec<u8>>> {
+                let buf = rmp_serde::to_vec(&op)?;
+                v.push(buf);
+                Ok(v)
+            })?;
+
+        self.update_mp(key, &mp_encoded_ops)
+    }
+
+    pub fn update_mp<K>(&mut self, key: &K, ops: &[Vec<u8>]) -> Result<Option<Tuple>, Error>
+        where
+            K: AsTuple,
     {
         let key_buf = key.serialize_as_tuple().unwrap();
         let key_buf_ptr = key_buf.as_ptr() as *const c_char;
-        let ops_buf = ops.serialize_as_tuple().unwrap();
+        let mut buf = Vec::with_capacity(128);
+        rmp::encode::write_array_len(&mut buf, ops.len() as u32)?;
+        ops.iter().try_for_each(|op_buf| buf.write_all(op_buf))?;
+        let ops_buf = unsafe { TupleBuffer::from_vec_unchecked(buf) };
         let ops_buf_ptr = ops_buf.as_ptr() as *const c_char;
         tuple_from_box_api!(
             ffi::box_update[
                 self.space_id,
                 self.index_id,
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
                 ops_buf_ptr,
-                ops_buf_ptr.offset(ops_buf.len() as isize),
+                ops_buf_ptr.add(ops_buf.len()),
                 0,
                 @out
             ]
@@ -339,10 +558,8 @@ impl Index {
     /// - `value` - encoded tuple in MsgPack Array format (`[field1, field2, ...]`)
     /// - `ops` - encoded operations in MsgPack array format, e.g. `[['=', field_id, value], ['!', 2, 'xxx']]`
     ///
-    /// Returns a new tuple.
-    ///
     /// See also: [index.update()](#method.update)
-    pub fn upsert<T, Op>(&mut self, value: &T, ops: &Vec<Op>) -> Result<Option<Tuple>, Error>
+    pub fn upsert<T, Op>(&mut self, value: &T, ops: &[Op]) -> Result<(), Error>
     where
         T: AsTuple,
         Op: AsTuple,
@@ -356,13 +573,16 @@ impl Index {
                 self.space_id,
                 self.index_id,
                 value_buf_ptr,
-                value_buf_ptr.offset(value_buf.len() as isize),
+                value_buf_ptr.add(value_buf.len()),
                 ops_buf_ptr,
-                ops_buf_ptr.offset(ops_buf.len() as isize),
+                ops_buf_ptr.add(ops_buf.len()),
                 0,
                 @out
             ]
         )
+            .map(|t| if t.is_some() {
+                unreachable!("Upsert doesn't return a tuple")
+            })
     }
 
     /// Return the number of elements in the index.
@@ -376,6 +596,11 @@ impl Index {
         }
     }
 
+    #[inline(always)]
+    pub fn is_empty(&self) -> Result<bool, Error> {
+        self.len().map(|l| l == 0)
+    }
+
     /// Return the number of bytes used in memory by the index.
     pub fn bsize(&self) -> Result<usize, Error> {
         let result = unsafe { ffi::box_index_bsize(self.space_id, self.index_id) };
@@ -417,7 +642,7 @@ impl Index {
                 self.space_id,
                 self.index_id,
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
                 @out
             ]
         )
@@ -439,7 +664,7 @@ impl Index {
                 self.space_id,
                 self.index_id,
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
                 @out
             ]
         )
@@ -462,7 +687,7 @@ impl Index {
                 self.index_id,
                 iterator_type.to_i32().unwrap(),
                 key_buf_ptr,
-                key_buf_ptr.offset(key_buf.len() as isize),
+                key_buf_ptr.add(key_buf.len()),
             )
         };
 
