@@ -9,13 +9,14 @@
 use std::fmt;
 use std::os::raw::c_char;
 
+use num_derive::ToPrimitive;
 use num_traits::ToPrimitive;
 use serde::{Serialize, Serializer};
 use serde_json::{Map, Value};
 
 use crate::error::{Error, TarantoolError};
 use crate::ffi::tarantool as ffi;
-use crate::index::{Index, IndexIterator, IteratorType};
+use crate::index::{self, Index, IndexIterator, IteratorType};
 use crate::tuple::{AsTuple, Tuple};
 use crate::tuple_from_box_api;
 
@@ -30,7 +31,7 @@ pub const SYSTEM_ID_MAX: u32 = 511;
 /// let schema_space = SystemSpace::Schema.into();
 /// ```
 #[repr(u32)]
-#[derive(Debug, Clone, PartialEq, ToPrimitive)]
+#[derive(Clone, Debug, PartialEq, Eq, ToPrimitive)]
 pub enum SystemSpace {
     /// Space if of _vinyl_deferred_delete.
     VinylDeferredDelete = 257,
@@ -84,16 +85,16 @@ pub enum SystemSpace {
     SessionSettings = 380,
 }
 
-impl Into<Space> for SystemSpace {
-    fn into(self) -> Space {
+impl From<SystemSpace> for Space {
+    fn from(ss: SystemSpace) -> Self {
         Space {
-            id: self.to_u32().unwrap(),
+            id: ss.to_u32().unwrap(),
         }
     }
 }
 
 /// Type of engine, used by space.
-#[derive(Copy, Clone, Debug)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq)]
 pub enum SpaceEngineType {
     Memtx,
     Vinyl,
@@ -115,26 +116,26 @@ impl Serialize for SpaceEngineType {
 /// (for details see [Options for box.schema.space.create](https://www.tarantool.io/en/doc/latest/reference/reference_lua/box_schema/space_create/)).
 ///
 /// `format` option is not supported at this moment.
-#[derive(Serialize)]
+#[derive(Clone, Debug, Serialize)]
 pub struct SpaceCreateOptions {
     pub if_not_exists: bool,
-    pub engine: Option<SpaceEngineType>,
+    pub engine: SpaceEngineType,
     pub id: Option<u32>,
-    pub field_count: Option<u32>,
+    pub field_count: u32,
     pub user: Option<String>,
     pub is_local: bool,
     pub is_temporary: bool,
     pub is_sync: bool,
-    pub format: Option<Vec<SpaceFieldFormat>>,
+    pub format: Option<Vec<Field>>,
 }
 
 impl Default for SpaceCreateOptions {
     fn default() -> Self {
         SpaceCreateOptions {
             if_not_exists: false,
-            engine: None,
+            engine: SpaceEngineType::Memtx,
             id: None,
-            field_count: None,
+            field_count: 0,
             user: None,
             is_local: true,
             is_temporary: true,
@@ -144,19 +145,75 @@ impl Default for SpaceCreateOptions {
     }
 }
 
+////////////////////////////////////////////////////////////////////////////////
+// Field
+////////////////////////////////////////////////////////////////////////////////
+
+#[deprecated = "Use `space::Field` instead"]
+pub type SpaceFieldFormat = Field;
+
 #[derive(Clone, Debug, Serialize)]
-pub struct SpaceFieldFormat {
-    pub name: String,
+pub struct Field {
+    pub name: String, // TODO(gmoshkin): &str
     #[serde(alias = "type")]
     pub field_type: SpaceFieldType,
+    pub is_nullable: bool,
 }
 
-impl SpaceFieldFormat {
+macro_rules! define_constructors {
+    ($($constructor:ident ($type:path))+) => {
+        $(
+            #[doc = ::std::concat!(
+                "Create a new field format specifier with the given `name` and ",
+                "type \"", ::std::stringify!($constructor), "\""
+            )]
+            pub fn $constructor(name: &str) -> Self {
+                Self {
+                    name: name.into(),
+                    field_type: $type,
+                    is_nullable: false,
+                }
+            }
+        )+
+    }
+}
+
+impl Field {
+    #[deprecated = "Use one of `Field::any`, `Field::unsigned`, `Field::string`, etc. instead"]
+    /// Create a new field format specifier.
+    ///
+    /// You should use one of the other constructors instead
     pub fn new(name: &str, ft: SpaceFieldType) -> Self {
-        return SpaceFieldFormat {
+        Self {
             name: name.to_string(),
             field_type: ft,
-        };
+            is_nullable: false,
+        }
+    }
+
+    /// Specify if the current field can be nullable or not. This method
+    /// captures `self` by value and returns it, so it should be used in a
+    /// builder fashion.
+    /// ```rust
+    /// let f = Field::string("middle name").is_nullable(true);
+    /// ```
+    pub fn is_nullable(mut self, is_nullable: bool) -> Self {
+        self.is_nullable = is_nullable;
+        self
+    }
+
+    define_constructors!{
+        any(SpaceFieldType::Any)
+        unsigned(SpaceFieldType::Unsigned)
+        string(SpaceFieldType::String)
+        number(SpaceFieldType::Number)
+        double(SpaceFieldType::Double)
+        integer(SpaceFieldType::Integer)
+        boolean(SpaceFieldType::Boolean)
+        decimal(SpaceFieldType::Decimal)
+        uuid(SpaceFieldType::Uuid)
+        array(SpaceFieldType::Array)
+        scalar(SpaceFieldType::Scalar)
     }
 }
 
@@ -175,13 +232,31 @@ pub enum SpaceFieldType {
     Scalar,
 }
 
-impl fmt::Display for SpaceFieldType {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{:?}", self)
+impl SpaceFieldType {
+    pub fn as_str(&self) -> &'static str {
+        match self {
+            Self::Any => "any",
+            Self::Unsigned => "unsigned",
+            Self::String => "string",
+            Self::Number => "number",
+            Self::Double => "double",
+            Self::Integer => "integer",
+            Self::Boolean => "boolean",
+            Self::Decimal => "decimal",
+            Self::Uuid => "uuid",
+            Self::Array => "array",
+            Self::Scalar => "scalar",
+        }
     }
 }
 
-#[derive(Serialize, Debug)]
+impl fmt::Display for SpaceFieldType {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str(self.as_str())
+    }
+}
+
+#[derive(Clone, Debug, Serialize)]
 pub struct FuncMetadata {
     pub id: u32,
     pub owner: u32,
@@ -206,7 +281,7 @@ pub struct FuncMetadata {
 
 impl AsTuple for FuncMetadata {}
 
-#[derive(Serialize, Debug)]
+#[derive(Clone, Debug, Serialize)]
 pub struct Privilege {
     pub grantor: u32,
     pub grantee: u32,
@@ -217,11 +292,19 @@ pub struct Privilege {
 
 impl AsTuple for Privilege {}
 
+#[derive(Clone, Debug)]
 pub struct Space {
     id: u32,
 }
 
 impl Space {
+    /// Return a space builder.
+    ///
+    /// - `name` - name of space to be created
+    pub fn builder(name: &str) -> Builder {
+        Builder::new(name)
+    }
+
     /// Create a space.
     /// (for details see [box.schema.space.create()](https://www.tarantool.io/en/doc/latest/reference/reference_lua/box_schema/space_create/)).
     ///
@@ -231,13 +314,13 @@ impl Space {
     /// Returns a new space.
     #[cfg(feature = "schema")]
     pub fn create(name: &str, opts: &SpaceCreateOptions) -> Result<Space, Error> {
-        return crate::schema::space::create_space(name, opts);
+        crate::schema::space::create_space(name, opts)
     }
 
     /// Drop a space.
     #[cfg(feature = "schema")]
     pub fn drop(&self) -> Result<(), Error> {
-        return crate::schema::space::drop_space(self.id);
+        crate::schema::space::drop_space(self.id)
     }
 
     /// Find space by name.
@@ -274,8 +357,15 @@ impl Space {
         name: &str,
         opts: &crate::index::IndexOptions,
     ) -> Result<Index, Error> {
-        crate::schema::index::create_index(self.id, name, opts)?;
-        Ok(self.index(name).unwrap())
+        crate::schema::index::create_index(self.id, name, opts)
+    }
+
+    /// Return an index builder.
+    ///
+    /// - `name` - name of index to create, which should conform to the rules for object names.
+    #[cfg(feature = "schema")]
+    pub fn index_builder<'a>(&self, name: &'a str) -> index::Builder<'a> {
+        index::Builder::new(self.id, name)
     }
 
     /// Find index by name.
@@ -311,7 +401,7 @@ impl Space {
     /// Returns a new tuple.
     ///
     /// See also: `box.space[space_id]:insert(tuple)`
-    pub fn insert<T>(&mut self, value: &T) -> Result<Option<Tuple>, Error>
+    pub fn insert<T>(&mut self, value: &T) -> Result<Tuple, Error>
     where
         T: AsTuple,
     {
@@ -321,10 +411,11 @@ impl Space {
             ffi::box_insert[
                 self.id,
                 buf_ptr,
-                buf_ptr.offset(buf.len() as isize),
+                buf_ptr.add(buf.len()),
                 @out
             ]
         )
+            .map(|t| t.expect("Returned tuple cannot be null"))
     }
 
     /// Insert a tuple into a space.
@@ -336,7 +427,7 @@ impl Space {
     /// - `value` - tuple value to replace with
     ///
     /// Returns a new tuple.
-    pub fn replace<T>(&mut self, value: &T) -> Result<Option<Tuple>, Error>
+    pub fn replace<T>(&mut self, value: &T) -> Result<Tuple, Error>
     where
         T: AsTuple,
     {
@@ -346,16 +437,17 @@ impl Space {
             ffi::box_replace[
                 self.id,
                 buf_ptr,
-                buf_ptr.offset(buf.len() as isize),
+                buf_ptr.add(buf.len()),
                 @out
             ]
         )
+            .map(|t| t.expect("Returned tuple cannot be null"))
     }
 
     /// Insert a tuple into a space. If a tuple with the same primary key already exists, replaces the existing tuple
     /// with a new one. Alias for [space.replace()](#method.replace)
     #[inline(always)]
-    pub fn put<T>(&mut self, value: &T) -> Result<Option<Tuple>, Error>
+    pub fn put<T>(&mut self, value: &T) -> Result<Tuple, Error>
     where
         T: AsTuple,
     {
@@ -379,6 +471,11 @@ impl Space {
         self.primary_key().len()
     }
 
+    #[inline(always)]
+    pub fn is_empty(&self) -> Result<bool, Error> {
+        self.len().map(|l| l == 0)
+    }
+
     /// Number of bytes in the space.
     ///
     /// This number, which is stored in Tarantool’s internal memory, represents the total number of bytes in all tuples,
@@ -453,7 +550,7 @@ impl Space {
     ///
     /// See also: [space.upsert()](#method.upsert)
     #[inline(always)]
-    pub fn update<K, Op>(&mut self, key: &K, ops: &Vec<Op>) -> Result<Option<Tuple>, Error>
+    pub fn update<K, Op>(&mut self, key: &K, ops: &[Op]) -> Result<Option<Tuple>, Error>
     where
         K: AsTuple,
         Op: AsTuple,
@@ -461,6 +558,22 @@ impl Space {
         self.primary_key().update(key, ops)
     }
 
+    /// Update a tuple using `ops` already encoded in message pack format.
+    ///
+    /// This function is similar to [`update`](#method.update) but instead
+    /// of a generic type parameter `Op` it accepts preencoded message pack
+    /// values. This is usefull when the operations have values of different
+    /// types.
+    ///
+    /// Returns a new tuple.
+    #[inline(always)]
+    pub fn update_mp<K>(&mut self, key: &K, ops: &[Vec<u8>]) -> Result<Option<Tuple>, Error>
+    where
+        K: AsTuple,
+    {
+        self.primary_key().update_mp(key, ops)
+    }
+
     /// Update or insert a tuple.
     ///
     /// If there is an existing tuple which matches the key fields of tuple, then the request has the same effect as
@@ -474,11 +587,9 @@ impl Space {
     /// - `value` - encoded tuple in MsgPack Array format (`[field1, field2, ...]`)
     /// - `ops` - encoded operations in MsgPack array format, e.g. `[['=', field_id, value], ['!', 2, 'xxx']]`
     ///
-    /// Returns a new tuple.
-    ///
     /// See also: [space.update()](#method.update)
     #[inline(always)]
-    pub fn upsert<T, Op>(&mut self, value: &T, ops: &Vec<Op>) -> Result<Option<Tuple>, Error>
+    pub fn upsert<T, Op>(&mut self, value: &T, ops: &[Op]) -> Result<(), Error>
     where
         T: AsTuple,
         Op: AsTuple,
@@ -486,3 +597,90 @@ impl Space {
         self.primary_key().upsert(value, ops)
     }
 }
+
+////////////////////////////////////////////////////////////////////////////////
+// Builder
+////////////////////////////////////////////////////////////////////////////////
+
+pub struct Builder<'a> {
+    name: &'a str,
+    opts: SpaceCreateOptions,
+}
+
+macro_rules! define_setters {
+    ($( $setter:ident ( $field:ident : $ty:ty ) )+) => {
+        $(
+            #[inline(always)]
+            pub fn $setter(mut self, $field: $ty) -> Self {
+                self.opts.$field = $field.into();
+                self
+            }
+        )+
+    }
+}
+
+impl<'a> Builder<'a> {
+    pub fn new(name: &'a str) -> Self {
+        Self {
+            name,
+            opts: Default::default(),
+        }
+    }
+
+    define_setters!{
+        if_not_exists(if_not_exists: bool)
+        engine(engine: SpaceEngineType)
+        id(id: u32)
+        field_count(field_count: u32)
+        user(user: String)
+        is_local(is_local: bool)
+        is_temporary(is_temporary: bool)
+        is_sync(is_sync: bool)
+        format(format: Vec<Field>)
+    }
+
+    pub fn field(mut self, field: Field) -> Self {
+        self.opts.format.get_or_insert_with(|| Vec::with_capacity(16))
+            .push(field);
+        self
+    }
+
+    #[cfg(feature = "schema")]
+    pub fn create(self) -> crate::Result<Space> {
+        crate::schema::space::create_space(self.name, &self.opts)
+    }
+}
+
+////////////////////////////////////////////////////////////////////////////////
+// macros
+////////////////////////////////////////////////////////////////////////////////
+
+/// Update a tuple or index.
+///
+/// The helper macro with semantic same as `space.update()`/`index.update()` functions, but supports
+/// different types in `ops` argument.
+///
+/// - `target` - updated space or index.
+/// - `key` - encoded key in MsgPack Array format (`[part1, part2, ...]`).
+/// - `ops` - encoded operations in MsgPack array format, e.g. `[['=', field_id, 100], ['!', 2, 'xxx']]`
+///
+/// Returns a new tuple.
+///
+/// See also: [space.update()](#method.update)
+#[macro_export]
+macro_rules! update {
+        ($target:expr, $key: expr, $($op:expr),+ $(,)?) => {
+            {
+                use std::borrow::Borrow;
+                let mut f = || -> $crate::Result<::std::option::Option<$crate::tuple::Tuple>> {
+                    let ops = [
+                        $(
+                            $crate::util::rmp_to_vec($op.borrow())?,
+                        )+
+                    ];
+                    $target.update_mp($key.borrow(), &ops)
+                };
+                f()
+            }
+        };
+    }
